#include "stdafx.h"
#include "MachineState.h"
#include "MemorySystem.h"

using namespace Simulator;
using namespace System::Globalization;
using namespace System::IO;
using namespace System::Windows::Forms;

MachineState::MachineState(String^ programRomPath, String^ microRomPath, unsigned long loadAddress, unsigned long bootAddress) :
	m_memorySystem(gcnew MemorySystem(programRomPath, loadAddress)),
	m_microrom(gcnew array<unsigned long>(s_microromSize)),
	m_startAddress(bootAddress)
{
	for (unsigned long addr = 0; addr<s_microromSize; addr++)
	{
		m_microrom[addr] = 0;
	}

	for (int i=0; i<3; i++)
	{
		if (!LoadMicrorom(microRomPath, i))
		{
			Application::Exit();
			return;
		}
	}

	Reset();
}

// Load a microrom image
bool MachineState::LoadMicrorom(String^ romPath, int romIndex)
{
	bool result = false;
	
	try 
	{
		StreamReader^ stream = File::OpenText(romPath + String::Format("\\microcode_rom_{0}.rom", romIndex));

		String^ data = stream->ReadToEnd();
		unsigned long len = data->Length;
		unsigned long addr = 0;
		unsigned long offset = 0;
		
		// parse each hex byte 
		while (offset < len)
		{
			Byte value;
			if (Byte::TryParse(data->Substring(offset, 2), NumberStyles::HexNumber, NumberFormatInfo::InvariantInfo, value))
			{
				m_microrom[addr] |= value << (romIndex * 8);
				addr++;
			}

			offset += 2;

			while (offset < len && Char::IsWhiteSpace(data, offset))
			{
				offset++;
			}
		}

		result = true;
	}
	catch (Exception^)
	{
		MessageBox::Show(String::Format("Could not load microcode_rom_{0}.rom", romIndex));
	}

	return result;
}

// Reset the machine to the initial state
void MachineState::Reset()
{
	Phase = 0;
	RegPC = m_startAddress + 1;
	RegSP = m_startAddress | 0xFFFF;
	RegAR = m_startAddress;
	RegOpcode = m_memorySystem->Load(m_startAddress);
	FlagNotInterruptEnable = true;
	FlagNotIrq = true;
	m_memorySystem->ClearLCD();
}

// Get a word of microinstruction data
unsigned long MachineState::GetMicroinstruction(unsigned long microAddress)
{
	return m_microrom[microAddress];
}

// Compute the current micro address
unsigned long MachineState::GetCurrentMicroaddress()
{
	unsigned long microAddress = RegOpcode * 512 + Phase * 32;

	if (FlagNotOverflow)
	{
		microAddress += 16;
	}
	if (FlagNotZero)
	{
		microAddress += 8;
	}
	if (FlagNegative)
	{
		microAddress += 4;
	}
	if (FlagNotCarry)
	{
		microAddress += 2;
	}
	if (FlagNotAddressCarry)
	{
		microAddress += 1;
	}

	return microAddress;
}

// Get the current microinstruction data
unsigned long MachineState::GetCurrentMicroinstruction()
{
	return m_microrom[GetCurrentMicroaddress()];
}

// "Peek" at a byte of data without changing any machine state
unsigned char MachineState::Peek(unsigned long address)
{ 
	return m_memorySystem->Peek(address); 
}

// Store a byte in RAM or HW device
void MachineState::Store(unsigned long address, unsigned char value)
{ 
	m_memorySystem->Store(address, value); 
}

// Get the LCD text contained within the memory system
String^ MachineState::LCDText() 
{ 
	return m_memorySystem->LCDText(); 
}

// Extract the value of an individual field from a bitfield
unsigned long MachineState::FieldValue(unsigned long bitfield, BitPos pos, int len)
{
	// compute mask of 2^len - 1
	unsigned long mask = 1;
	for (int i=0; i<len; i++)
	{
		mask *= 2;
	}
	mask = mask - 1;

	return (bitfield >> ((int)pos - len + 1)) & mask;
}

// Simulate one clock cycle, executing one micro-instruction
unsigned long MachineState::MicroStep()
{
	unsigned long addressModified = 0xFFFFFFFE;

	unsigned long microOp = GetCurrentMicroinstruction();

	// effective address
	unsigned long effectiveAddress = RegPC;
	unsigned long addressSelect = FieldValue(microOp, BitPos::kAddressSelect, 2);
	if ((unsigned long)AddressSelect::kSelectSP == addressSelect)
	{
		effectiveAddress = RegSP;
	}
	else if ((unsigned long)AddressSelect::kSelectAR == addressSelect)
	{
		effectiveAddress = RegAR;
	}

	// ALU
	bool carryOut, overflowOut;
	unsigned char aluResult = RunALU(microOp, effectiveAddress, carryOut, overflowOut);

	// address carry
	if (IsBitZero(microOp, BitPos::kDest3))
	{
		FlagNotAddressCarry = carryOut;
	}

	// condition codes
	unsigned long s0 = FieldValue(microOp, BitPos::kS0, 1);
	unsigned long s1 = FieldValue(microOp, BitPos::kS1, 1);
	if (s1 && s0)
	{
		// load all
		FlagNotCarry = carryOut;
		FlagNotZero = (aluResult != 0);
		FlagNegative = (aluResult & 0x80) != 0; 

		// load overflow only when op is A <- ? * T
		unsigned long dataDest = FieldValue(microOp, BitPos::kDest, 3);
		if (IsBitOne(microOp, BitPos::kDest3))
		{
			dataDest += 8;
		}
		unsigned long rightOp = FieldValue(microOp, BitPos::kALURight, 2);
		
		if ((unsigned long)Dest::kLoadRegA == dataDest &&
			(unsigned long)ALURight::kRightRegT == rightOp)
		{
			FlagNotOverflow = overflowOut;
		}
	}
	else if (s1 && !s0)
	{
		// shift right
		FlagNotCarry = FlagNegative;
		FlagNegative = FlagNotZero;
		FlagNotZero = FlagNotOverflow;
		FlagNotOverflow = (RegX & 0x8) != 0;
	}
	else if (!s1 && s0)
	{
		// load N and Z
		FlagNotZero = (aluResult != 0);
		FlagNegative = (aluResult & 0x80) != 0;
	}

	// move data
	unsigned char dataValue;
	
	if (IsBitZero(microOp, BitPos::kDriveALU))
	{
		dataValue = aluResult;
	}
	else
	{
		dataValue = m_memorySystem->Load(effectiveAddress);
		if (effectiveAddress == 0x3F13 && !FlagNotIrq && dataValue == 1) 
		{
			// clear IRQ after reading keyboard control memory-mapped location	
			FlagNotIrq = true;
		}
	}

	unsigned long dataDest = FieldValue(microOp, BitPos::kDest, 3);
	if (IsBitOne(microOp, BitPos::kDest3))
	{
		dataDest += 8;
	}
	// determine the move destination
	if ((unsigned long)Dest::kLoadSPHI == dataDest)
	{
		RegSP &= 0x00FF00FF;
		RegSP |= (dataValue << 8);
	}
	else if ((unsigned long)Dest::kLoadSPBank == dataDest)
	{
		RegSP &= 0x0000FFFF;
		RegSP |= (dataValue << 16);
	}
	else if ((unsigned long)Dest::kLoadARLO == dataDest)
	{
		RegAR &= 0x00FFFF00;
		RegAR |= (dataValue);
	}
	else if ((unsigned long)Dest::kLoadARHI == dataDest)
	{
		RegAR &= 0x00FF00FF;
		RegAR |= (dataValue << 8);
	}
	else if ((unsigned long)Dest::kLoadARBank == dataDest)
	{
		RegAR &= 0x0000FFFF;
		RegAR |= (dataValue << 16);
	}
	else if ((unsigned long)Dest::kLoadPCLO == dataDest)
	{
		RegPC &= 0x00FFFF00;
		RegPC |= (dataValue);
	}
	else if ((unsigned long)Dest::kLoadPCHI == dataDest)
	{
		RegPC &= 0x00FF00FF;
		RegPC |= (dataValue << 8);
	}
	else if ((unsigned long)Dest::kLoadPCBank == dataDest)
	{
		RegPC &= 0x0000FFFF;
		RegPC |= (dataValue << 16);
	}
	else if ((unsigned long)Dest::kLoadMem == dataDest)
	{
		m_memorySystem->Store(effectiveAddress, dataValue);
		addressModified = effectiveAddress;
	}
	else if ((unsigned long)Dest::kLoadSPLO == dataDest)
	{
		RegSP &= 0x00FFFF00;
		RegSP |= (dataValue);
	}
	else if ((unsigned long)Dest::kLoadRegT == dataDest)
	{
		RegT = dataValue;
	}
	else if ((unsigned long)Dest::kLoadRegY == dataDest)
	{
		RegY = dataValue;
	}
	else if ((unsigned long)Dest::kLoadRegX == dataDest)
	{
		RegX = dataValue;
	}
	else if ((unsigned long)Dest::kLoadRegA == dataDest)
	{
		RegA = dataValue;
	}

	// phase and opcode
	if (IsBitZero(microOp, BitPos::kLoadOp))
	{
		Phase = 0;
		
		if (!FlagNotIrq && !FlagNotInterruptEnable)
		{
			RegOpcode = 0;
		}
		else
		{
			RegOpcode = m_memorySystem->Load(effectiveAddress);
		}
	}
	else
	{
		Phase = (Phase + 1) % 16;
	}

	// special
	unsigned long special = FieldValue(microOp, BitPos::kSpecial, 2);
	if ((unsigned long)Special::kLoadIE == special)
	{
		if (IsBitZero(microOp, BitPos::kSPDirIEData))
		{
			FlagNotInterruptEnable = false;
		}
		else
		{
			FlagNotInterruptEnable = true;
		}
	}
	else if ((unsigned long)Special::kIncAR == special)
	{
		// increment and roll over within 24-bit range
		RegAR++;
		RegAR = RegAR & 0x00FFFFFF;
	}
	else if ((unsigned long)Special::kIncPC == special)
	{
		// increment and roll over within 24-bit range
		RegPC++;
		RegPC = RegPC & 0x00FFFFFF;
	}

	if (IsBitZero(microOp, BitPos::kCountSP))
	{
		if (IsBitOne(microOp, BitPos::kSPDirIEData))
		{
			RegSP++;
		}
		else
		{
			RegSP--;
		}

		RegSP = RegSP & 0x00FFFFFF;
	}

	return addressModified;
}

// simulate the behavior of a pair of 74LS181's
unsigned char MachineState::RunALU(unsigned long microOp, unsigned long effectiveAddress, bool% carryOut, bool% overflowOut)
{
	unsigned char a;
	unsigned char b;
	unsigned char result;
	long tempResult;

	// left operand
	unsigned long leftOp = FieldValue(microOp, BitPos::kALULeft, 2);
	if ((unsigned long)ALULeft::kLeftZero == leftOp)
	{
		a = 0;
	}
	else if ((unsigned long)ALULeft::kLeftRegX == leftOp)
	{
		a = RegX;
	}
	else if ((unsigned long)ALULeft::kLeftRegA == leftOp)
	{
		a = RegA;
	}
	else if ((unsigned long)ALULeft::kLeftRegYorX7 == leftOp)
	{
		// y and x7 share a control bit: use x7 with address register destinations, otherwise use Y
		unsigned long dest3 = FieldValue(microOp, BitPos::kDest3, 1);
		if (dest3)
		{
			a = RegY;
		}
		else
		{
			a = RegX & 0x80 ? 0xFF : 0;
		}
	}

	// right operand
	unsigned long rightOp = FieldValue(microOp, BitPos::kALURight, 2);
	if ((unsigned long)ALURight::kRightAddressLo == rightOp)
	{
		b = (unsigned char)(effectiveAddress & 0xFF);
	}
	else if ((unsigned long)ALURight::kRightAddressHi == rightOp)
	{
		b = (unsigned char)((effectiveAddress & 0xFF00) >> 8);
	}
	else if ((unsigned long)ALURight::kRightAddressBank == rightOp)
	{
		b = (unsigned char)((effectiveAddress & 0xFF0000) >> 16);
	}
	else if ((unsigned long)ALURight::kRightRegT == rightOp)
	{
		b = RegT;
	}

	unsigned char c = (unsigned char)FieldValue(microOp, BitPos::kALUCarry, 1);
	unsigned long function = FieldValue(microOp, BitPos::kALUOp, 4);

	if (IsBitOne(microOp, BitPos::kALUMode))
	{
		// logical mode
		switch (function)
		{
		case 0:
			result = ~a;
			break;
		case 1:
			result = ~(a|b);
			break;
		case 2:
			result = (~a)&b;
			break;
		case 3:
			result = 0;
			break;
		case 4:
			result = ~(a&b);
			break;
		case 5:
			result = ~b;
			break;
		case 6:
			result = a^b;
			break;
		case 7:
			result = a^(~b);
			break;
		case 8:
			result = (~a)|b;
			break;
		case 9:
			result = ~(a^b);
			break;
		case 10:
			result = b;
			break;
		case 11:
			result = a&b;
			break;
		case 12:
			result = 0xFF;
			break;
		case 13:
			result = a|(~b);
			break;
		case 14:
			result = a|b;
			break;
		case 15:
			result = a;
			break;
		}

		// carryOut and overflowOut aren't defined for logical mode
		carryOut = false;
		overflowOut = false;
	}
	else
	{
		// arithmetic mode
		switch (function)
		{
		case 0:
			tempResult = (long)a + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 1:
			tempResult = (long)(a|b) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 2:
			tempResult = (long)(a|(~b)) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 3:
			tempResult = -(long)c;
			carryOut = tempResult < 0;
			break;
		case 4:
			tempResult = (long)a + (a&(~b)) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 5:
			tempResult = (long)(a|b) + (a&(~b)) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 6:
			tempResult = (long)a - b - c;
			carryOut = tempResult < 0;
			break;
		case 7:
			tempResult = (long)(a&(~b)) - c;
			carryOut = tempResult < 0;
			break;
		case 8:
			tempResult = (long)a + (a&b) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 9:
			tempResult = (long)a + b + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 10:
			tempResult = (long)(a|b) + (a&b) + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 11:
			tempResult = (long)(a&b) - c;
			carryOut = tempResult < 0;
			break;
		case 12:
			tempResult = (long)a + a + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 13:
			tempResult = (long)(a|b) + a + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 14:
			tempResult = (long)(a|(~b)) + a + 1 - c;
			carryOut = tempResult < 256;
			break;
		case 15:
			tempResult = (long)a - c;
			carryOut = tempResult < 0;
			break;
		}

		result = (unsigned char)(tempResult & 0xFF);

		//assign overflow_bar = ~ ((~A[7] & ~B[7] & result[7] & ~func[1]) | (A[7] & B[7] & ~result[7] & ~func[1]) | (~A[7] & B[7] & result[7] & func[1]) | (A[7] & ~B[7] & ~result[7] & func[1]));
		bool a7 = (a & 0x80) != 0;
		bool b7 = (b & 0x80) != 0;
		bool r7 = (result & 0x80) != 0;
		bool f1 = (function & 0x2) != 0;
		overflowOut = !((!a7 && !b7 && r7 && !f1) || (a7 && b7 && !r7 && !f1) || (!a7 && b7 && r7 && f1) || (a7 && !b7 && !r7 && f1));
	}

	return result;
}